#!/usr/bin/env python
import os
import re
import xmlrpclib
import sys
import time
import socket
import uu
import StringIO
import xml.dom.minidom

class KestrelError(Exception):
  def __init__(self,msg):
    self.msg = "Kestrel Error: " + msg

  def writeSolutionFile(self,scrdir):
    fname = os.path.join(scrdir,'gamssolu.scr')
    if os.access(fname,os.W_OK):
      solution = open(fname,"w")
      solution.write("""  1 6.0000000000000000E+00
  2 1.3000000000000000E+01
  3 0.0000000000000000E+00
  4 0.0
  5 0.0000000000000000E+00
  6 0.0
  7 0.0
  8 0.0
  0 0.0\n""")
      solution.close()
    else:
      sys.stderr.write("Could not open %s\n" % fname)
    
  def __str__(self):
    return repr(self.msg)
  
class KestrelGamsClient:
  def __init__(self,argv,host="neos.mcs.anl.gov",port=3332):
    self.argv=argv
    self.scrdir="."
    self.neos = xmlrpclib.Server("http://%s:%d" % (host,port))

  def checkServer(self):
    self.xml = ""
    reply = self.neos.ping()
    if len(self.argv) >= 3:
      self.cntrfile = self.argv[2]
      self.scrdir = self.argv[1]
    else:
      raise KestrelError, "Usage: gamske_ux.out <scrdir> <cntrfile>"
    
    if reply.find('alive') < 0:
      raise KestrelError, "Unable to contact NEOS at http://%s:%d" % \
            (self.host, self.port)
      



  def createSubmission(self):
    try:
      allKestrelSolvers = self.neos.listSolversInCategory("kestrel")
      self.kestrelGamsSolvers = []
      for s in allKestrelSolvers:
        i = s.find(':GAMS')
        if i > 0:
          self.kestrelGamsSolvers.append(s[0:i])
        
      # Get the matrix, dictionary and instruction file
      gams_files = {}
      gams_files['cntr'] = self.getcntrfile()

      # Solver name is read from first line of 'kestrel.opt'
      solver_name = self.getSolverName()

      for key in ['matr','inst','dict']:
        fname=os.path.join(self.scrdir,"gams"+key+".scr")
        if os.access(fname,os.R_OK):
          f = open(fname,"r")
          val=""
          buf = f.read()
          while buf:
            val += buf
            buf=f.read()
          f.close()
          gams_files[key] = val

      self.xml = """ 
        <document>
        <category>kestrel</category>
        <solver>%s</solver>
        <inputType>GAMS</inputType>
        """ % (solver_name)

      for key in gams_files.keys():
        self.xml += "<%s><![CDATA[%s]]></%s>\n" % (key,gams_files[key],key)

      opt = file("kestrel.opt")
      self.xml += "<options><![CDATA[%s]]></options>" % opt.read()
      opt.close()

      self.xml += "</document>"

    except Exception, e:
      raise KestrelError, str(e)
    

  def submit(self):
    try:
      user = "%s on %s" % (os.getenv('LOGNAME'),
                           socket.getfqdn(socket.gethostname()))
      (self.jobNumber,self.password) = \
                         self.neos.submitJob(self.xml,user,"KESTREL")
      sys.stdout.write("job#=%d, pass=%s\n" % (self.jobNumber,self.password))
    except Exception, e:
      raise KestrelError, str(e)

  def getResults(self):
    try:
      status = "Waiting"
      offset = 0
      time.sleep(1)
      while (status == "Waiting" or status=="Running"):
        (results,offset) = self.neos.getIntermediateResults(self.jobNumber, self.password,offset)
        results = results.data
        if len(results):
          print results
        status = self.neos.getJobStatus(self.jobNumber,self.password)

      print "status=done"
      resultsXML = self.neos.getFinalResults(self.jobNumber,self.password)
      if isinstance(resultsXML,xmlrpclib.Binary):
        resultsXML = resultsXML.data
      self.parseSolution(resultsXML)

    except Exception, e:
      raise KestrelError, str(e)
    
  def getText(self,node):
    """
    Returns the text from the node of an xml document
    """
    s = ""
    if isinstance(node,(str,unicode)):
      return node
    if isinstance(node.nodeValue,(str,unicode)):
      return node.data
    elif node.hasChildNodes():
      for n in node.childNodes:
        s += self.getText(n)
    return s

  def getSolverName(self):
    solver_name = ""
    if os.access(self.optfilename,os.R_OK):
      optfile = open(self.optfilename,'r')
      for line in optfile:
        sys.stdout.write(line+"\n")
        m = re.match(r'kestrel_solver\s+(\S+)',line)
        if m:
          solver_name = m.groups()[0]
          break
      optfile.close()

      if solver_name and solver_name in self.kestrelGamsSolvers:
        return solver_name

      
      elif not solver_name:
        sys.stdout.write("No solver selected.  Choose from:\n")
        
      else:
        sys.stdout.write("%s is not available on NEOS.  Choose from:\n" % solver_name)
        
    else:
      sys.stdout.write("Could not read options file %s\n" % self.optfilename)
                       

    for s in self.kestrelGamsSolvers:
      sys.stdout.write("\t%s\n"%s)
    sys.stdout.write('To choose, add to file kestrel.opt:\nkestrel_solver <solver>\n')
    sys.exit(0)
    
  def parseSolution(self,xmlstring):
    try:
      doc = xml.dom.minidom.parseString(xmlstring)
      for tag in ['solu','log','stat']:
        node = doc.getElementsByTagName(tag)
        if node and len(node):
          f = open(os.path.join(self.scrdir,"gams%s.scr"%tag),'w')
          f.write(self.getText(node[0]))
          if tag=='log':
            sys.stdout.write(self.getText(node[0]))
          f.close()
      doc.unlink()
    except Exception, e:
      raise Exception, str(e)
      f = open(os.path.join(self.scrdir,'gamslog.scr'),'a')
      f.write(xmlstring)
      f.close()
      sys.stdout.write(xmlstring)
      f = open(os.path.join(self.scrdir,'gamssolu.scr'),'w')
      f.write("""  1 6.0000000000000000E+00
  2 1.3000000000000000E+01
  3 0.0000000000000000E+00
  4 0.0
  5 0.0000000000000000E+00
  6 0.0
  7 0.0
  8 0.0
  0 0.0\n""")

      
      

  def getcntrfile(self):
    """
    This function does the following with the cntr file
    line 13:
      extract is_ascii, use_options

    line 18,19,21,22,23,24
      remove absolute path from /dir/to/scrdir/gams*.scr

    line 20
      set options file to 'kestrel.scr'

    line 25
      set to '2' to write to log file

    line 28-30
      set working,system,scratch directories to '.'

    line 33,34,35
      remove licence

    line 37
      set parameter file

    line 38
      read #models #solvers
      ignore next #models + 2*#solvers with (SOLVER # # 0 ..) +
                 3*#solvers with (SOLVER # # 1 ...)   lines

    next two lines are more license (remove them)
  hh
    set directories of remaining paths to current directory
     (.scr, .so, sbbinfo.scr)
    """
    f = open(self.cntrfile,'r')
    lines = f.readlines()
    f.close()

    #extract is_ascii, use_options
    m = re.match(r'(\d+)\s+(\d+)',lines[12])
    if m and m.groups():
      self.is_ascii=m.groups()[0]
      self.use_options = m.groups()[1]
    else:
      raise IOError,("Bad line 13 in gamscntr.scr file.")

    # Get basename of files (make everything in local directory)
    for line in [17,18,20,21,22,23]:
      lines[line] = os.path.basename(lines[line])
      
    self.optfilename = lines[19].strip()
    
    lines[19] = "kestrel.opt\n"
    lines[24]="2\n" # '2' means write to log file
    self.scrdir = lines[29].strip()
    lines[27] = lines[28] = lines[29] = '.\n' # set working directories
    #remove license
    lines[32] = lines[33] = lines[34] = lines[-11] = lines[-10] = "\n"

    # Get basename of files (make everything in local directory)
    lines[-9] = 'model.scr\n'
    lines[-4] = 'model.so\n'
    lines[-3] = 'sbbinfo.scr\n'
    

    return "".join(lines[:37]) + "".join(lines[-11:])


if __name__=="__main__":
  kestrel = KestrelGamsClient(sys.argv)
  try:
    kestrel.checkServer()
    kestrel.createSubmission()
    kestrel.submit()
    kestrel.getResults()
  except KestrelError, (e):
    e.writeSolutionFile(kestrel.scrdir)
    sys.stdout.write(str(e)+"\n")
  
