#!/usr/bin/env python
import xmlrpclib
import logging
import logging.handlers
import socket
import SimpleXMLRPCServer
import SocketServer
import os
import sys
import signal
import time
import WebPrinter
import xml.dom.minidom

import NeosTools
from config import Variables

class NeosServer:
  def __init__(self):
    self.logger = logging.getLogger('NeosServer')
    hdlr = logging.handlers.RotatingFileHandler(os.path.join(Variables.LOGDIR,"NeosServer"),maxBytes=1000000,backupCount=2)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    self.logger.addHandler(hdlr) 
    self.logger.setLevel(logging.DEBUG)
    NeosServer.client_address = ""

  
  def help(self):
    import pydoc
    import re
    txt = pydoc.TextDoc()
    return re.sub(r'\x08.','',txt.docclass(NeosServer))

  def welcome(self):
    """Returns a welcome message.
    """
    return """\
    ____________________________________________________________________
    
                         WELCOME TO NEOS!
    
                 The Network-Enabled Optimization System

                            developed by
    
                 The Optimization Technology Center of
         Argonne National Laboratory and Northwestern University
    ____________________________________________________________________
    
    
    I. ABSTRACT
    
    This server accepts optimization problems sent electronically, and 
    solves them using the computing facilities of the Optimization 
    Technology Center and other participating hosts. The user can obtain
    information on the optimization problem, the solution process, and
    the solution. Further, users can register their own software with
    this server, making their code available to other users.
    
    This server is under construction.  Comments and suggestions should
    be sent to %s. 
    We have assumed familiarity with basic optimization terminology.  If 
    you need additional information, consult

    J. More' and S. Wright,
    Optimization Software Guide,
    SIAM Publications, 1993.



    II. DISCLAIMER

    This information is provided without any express or implied
    warranty.  In particular, there is no warranty of any kind 
    concerning the fitness of this information for any particular
    purpose.
    """    % (Variables.NEOS_COMMENTS)

  def emailHelp(self):
    """Returns general help message for email users
    """
    msg = self.welcome() + "\n\n"
    msg += "The following solvers are available on the NEOS server.\n"
    msg += "For more information on a particular solver, send:\n"
    msg += "help <category>:<solver>:<input>\n\n"
    for s in self.listAllSolvers():
      msg += "\t" + s + '\n'
    msg += "\n\n\nMore information on NEOS can be found at %s\n" \
           % Variables.NEOS_WEB_SITE
    return msg
  
  def version(self):
    """Returns version number of the NEOS server (string).
    """
    return "neos version 5"
     

  def ping(self):
    """Verifies that this NeosServer is running.

    Returns message 'NeosServer is alive'
    """
    return "NeosServer is alive\n"

  def pingHost(self,user,hostname):
    """Verifies a solver daemon.

    If there is a solver daemon running on this host
    under the user's account, then a confimation message
    is returned ('OK').  Otherwise, an error message is returned.
    """
    msg = ""
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      result = db.query('SELECT port FROM STATIONS WHERE user="%s" and host_name="%s"' %
                        (user,hostname))
      if result and len(result):
        port = int(result[0])
        host = xmlrpclib.Server("http://%s:%d" % (hostname,port))
        msg = host.ping()
      else:
        msg = "Unknown username and hostname: %s@%s" % (user,hostname)
      db.close()
    except Exception, e:
      self.logger.exception(str(e))
      msg = str(e)
      if db is not None:
        db.close()
      
    return msg

  def getSolverTemplate(self,category,solvername,inputMethod):
    """Returns a template for a given solver.

    If the solver category:solvername:inputMethod exists on NEOS, then
    an XML template is returned for using this solver via the XML-RPC
    or email interfaces.  
    """
    msg = ""
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg = solver.getTemplate(db,category,solvername,inputMethod)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return msg
      
  def printQueue(self):
    """Returns a string containing the current NEOS jobs.
    """
    msg =""
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      results = db.queryAll('SELECT QUEUE.job_number,JOBS.submit_time,JOBS.start_time,JOBS.solver_cat,JOBS.solver_name,JOBS.solver_input,STATIONS.host_name FROM QUEUE LEFT JOIN JOBS ON QUEUE.job_number=JOBS.number LEFT JOIN STATIONS ON STATIONS.ID=JOBS.station_id WHERE QUEUE.status="Running" ORDER BY QUEUE.job_number')
      msg += "Running:\n"
      if len(results) == 0:
        msg += "No Jobs Running.\n"
      for (job_number,submit_time,start_time,solver_cat,solver_name,solver_input,host_name) in results:
        msg += "job#%d\t%s:%s:%s\tsubmitted\t%s to %s\tstarted %s\n" % \
               (job_number,
                solver_cat,
                solver_name,
                solver_input,
                submit_time,
                host_name,
                start_time)

      results = db.queryAll('SELECT QUEUE.job_number,JOBS.submit_time,JOBS.solver_cat,JOBS.solver_name,JOBS.solver_input FROM QUEUE LEFT JOIN JOBS ON QUEUE.job_number=JOBS.number WHERE QUEUE.status="Waiting" ORDER BY QUEUE.job_number')
      msg += "\nQueued:\n"
      if len(results) == 0:
        msg += "No Jobs in Queue.\n"
      for (job_number, submit_time, solver_cat,solver_name,solver_input) in results:
        msg += "job#%d\t%s:%s:%s\tsubmitted %s\n" % (job_number,
                                                     solver_cat,
                                                     solver_name,
                                                     solver_input,
                                                     submit_time)

      msg += "\n"
      db.close()
    except Exception, (e):
      self.logger.exception("Print Queue:" + str(e))
      if db is not None:
        db.close()
      raise Exception, (e)
    return msg


  def submitJob(self,xmlstring,user="",interface=""):
    """Submit an optimization job to NEOS.

    Use this method to submit your job to NEOS.
    It will return a tuple containing (jobnumber,password).
    You can the use this jobnumber and password to get the
    results or status of your job using the methods getStatus,
    getIntermediateResults, and getFinalResults.

    For more information on the format of the xmlstring, you can
    use the getSolverTemplate function.
    """

    db = None
    if interface not in ['Web','Web Example','kestrel','email','JAVA Submission Tool']:
      interface = "other"
    # Check for user (don't trust what we're given, unless it comes
    # from neos server -- (via email or cgi interface)
    if NeosServer._client_address != socket.gethostbyname(Variables.NEOS_HOST):
      user = NeosServer._client_address
    
    #    if not user:
    #      user = self.client_address
    try:
      if isinstance(xmlstring,xmlrpclib.Binary):
        xmlstring = xmlstring.data
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(log=self.logger)
      (jobNumber,password) = job.submit(db,xmlstring,user,interface)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return (jobNumber, password)
      

  def killJob(self,jobNumber,password):
    """Cancel a NEOS submission.

    If you want to cancel a job submission running on NEOS (or
    waiting to run on NEOS), then use this method.  We require
    the job password to prevent abuse of this function.
    """

    msg = ""
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password)
      msg += job.kill(db)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      msg += str(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return msg

  def getJobStatus(self,jobNumber,password):
    """Get the current status of your job.

    Returns "Done", "Running", "Waiting", "Unknown Job", or "Bad Password".
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password)
      status = job.getJobStatus(db)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return status

  def getXML(self,category,name,input):
    """Returns the XML description of the given solver.

    This will be deprecated.  Use getSolverTemplate.
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      xmlstring = solver.getXML(db,category,name,input)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return xmlstring
  
  def listSolversOnStation(self,user,hostname):
    """List the solvers running on a given station.

    Goes through the SOLVERS table and determines which solvers
    are registered to run on this user@machine (plus pid,port).
    """
    list = []
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      results = db.queryAll('SELECT solver_cat,solver_name,solver_input,port,pid from STATIONS WHERE user="%s" AND host_name="%s"' % (user,hostname))
      for result in results:
        if result[3] is None or result[4] is None:
          list.append(":".join(result[0:3]))
        else:
          list.append("%s %d %d" % (":".join(result[0:3]), result[3], result[4]))

      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)

    return list

  def listAllSolvers(self):
    """Lists all solvers available on NEOS.

    Returns a list of category:solver:inputMethod
    """
    list = []
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      results = db.queryAll('SELECT category,name,input FROM SOLVERS WHERE status="Enabled"')
      db.close()
      for result in results:
        list.append(":".join(result))
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return list
    
  def listCategories(self):
    """Lists all categories available on NEOS

    Returns a dictionary with entries {'abbreviated name':'full name',...}
    If you want to provide a solver to NEOS in an unavailable category,
    please email the NEOS administrators and we will add it for you.
    """
    dict = {}
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      results = db.queryAll("SELECT abbreviation, full_name from CATEGORIES")
      db.close()
      for result in results:
        dict[result[0]] = result[1]
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return dict

    
  def listSolversInCategory(self,category):
    """List all NEOS solvers in this category
    
    Returns a list of solver:input for every solver in the category
    (category can be abbreviation or full_name)
    """
    list = []
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      results = db.queryAll('SELECT SOLVERS.name,SOLVERS.input FROM SOLVERS LEFT JOIN CATEGORIES ON SOLVERS.category=CATEGORIES.abbreviation WHERE status="Enabled" AND (CATEGORIES.abbreviation="%s" OR CATEGORIES.full_name="%s")' % (category, category))
      db.close()
      for result in results:
        list.append(':'.join(result))

    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return list

  def getFinalResults(self,jobNumber,password):
    """Gets results of job from NEOS.

    Retrieve results from NEOS.  If the job is still running, then
    this function will hang until the job is finished.

    This function will return a base-64 encoded object.
    Please read your XML-RPC client documentation for decoding.  

    (For Python's xmlrpclib library, you can use the object's 'data'
    data member).
    """
  
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password,log=self.logger)
      msg = job.getResults(db,'job_results',blocking=True)
      db.close()

    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)

    return xmlrpclib.Binary(msg)
    


  def getFinalResultsNonBlocking(self,jobNumber,password):
    """Gets results of job from NEOS (non-blocking).

    Retrieve results from NEOS.  If the job is still running, then
    this function will return an empty string (base-64 encoded).

    This function will return a base-64 encoded object.
    Please read your XML-RPC client documentation for decoding.  

    (For Python's xmlrpclib library, you can use the object's 'data'
    data member).
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password,log=self.logger)
      msg = job.getResults(db,'job_results',blocking=False)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return xmlrpclib.Binary(msg)
  
  def getIntermediateResults(self,jobNumber,password,offset):
    """Gets intermediate results of job from NEOS.

    Retrieve intermediate results from NEOS, starting at the character
    offset up to the last received data.  Intermediate results are usually
    the standard output of the solver daemon.
    
    If the job is still running, then this function will hang until
    another packet of output is sent to NEOS or the job is finished.
    The offset 

    This function will return a tuple of thebase-64 encoded object and
    the new offset.  The offset refers to uncoded characters.
    Please read your XML-RPC client documentation for decoding.  

    (For Python's xmlrpclib library, you can use the object's 'data'
    data member).
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password,log=self.logger)
      msg = job.getResults(db,'job_out',offset=offset,blocking=True)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return (xmlrpclib.Binary(msg),offset+len(msg))


  def getIntermediateResultsNonBlocking(self,jobNumber,password,off):
    """Gets intermediate results of job from NEOS.

    Retrieve intermediate results from NEOS.  Intermediate results are
    usually the standard output of the solver daemon.

    This function will return a tuple of thebase-64 encoded object and
    the new offset.  The offset refers to uncoded characters.
    Please read your XML-RPC client documentation for decoding.  

    (For Python's xmlrpclib library, you can use the object's 'data'
    data member).
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password,log=self.logger)
      msg = job.getResults(db,'job_out',offset=off,blocking=False)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return (xmlrpclib.Binary(msg),offset+len(msg))

  def registerDaemon(self,user,host,port, pid):
    """Register a solver daemon with NEOS.

    If you are maintaining a solver on the NEOS site, then this function
    will tell NEOS how to contact the solver daemon on your machine.
    """
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      db.query('UPDATE STATIONS SET port=%d,pid=%d WHERE host_name="%s" AND user="%s"'%(port,pid,host,user))
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return "Daemon registered.\n"
    
    
  def validateSolverXML(self,xmlString):
    """Checks the syntax of the solver XML description.

    Before submitting your solver description to NEOS, you can use
    this method to check if your syntax is correct.  It will either
    return 'OK' or an error message.
    """
    try:
      solver = NeosTools.NeosSolver(self.logger)
      solver.validateXml(xmlString)
    except Exception, (e):
      self.logger.exception(e)
      raise Exception, (e)
      
    return solver.validateXml(xmlString)

  def registerSolver(self,xmlString):
    """Register a new solver on NEOS

    This will create a new solver on NEOS.  Anyone is permitted
    to register their solver on NEOS.  If you would like to make
    your solver available, please read the documentation at
    http://neos.mcs.anl.gov/neos/neos-5/Web/SolverHowTo.html
    """ 
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      (ok,msg) = solver.register(db,xmlString)
      if ok:
        # Update web pages
        self.logger.debug("Updating web pages")
        WebPrinter.makeCatalog(db,self.logger)
        WebPrinter.makeTestCatalog(db,self.logger)
        WebPrinter.makeWebPages(xmlString,self.logger)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return msg

  def registerExample(self,xmlstring,password):
    """Register an example on the NEOS web pages.

    This function will upload an example onto the NEOS web page
    for your solver.
    """
    msg = ""
    solver={}
    db = None
    try:
      db= NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg = solver.registerExample(db,xmlstring,password)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
     
    return msg

  def removeExample(self,category,solvername,input,passwd,examplename):
    """Remove an example from NEOS

    This will remove the example from your NEOS solver page.
    """
    db = None
    msg = ""
    try:
      db = NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg=solver.removeExample(db,category,solvername,input,passwd,examplename)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, e
    
    return msg

  def disableSolver(self,category,solvername,input,password):
    """Temporarily disable your solver.

    If you wish to temporarily remove your solver from NEOS, then you
    can use this method.  Your solver will be removed from the NEOS solvers
    index, and any users who attempt to use your solver while
    disabled will receive a message that it has been disabled.
    Use 'enableSolver' to reenable.
    """
    
    db = None
    try:
      db= NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg = solver.disable(db,category,solvername,input,password)

      # Update web pages
      if msg.find('now disabled') >= 0:
        WebPrinter.makeCatalog(db,self.logger)

      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg

  def enableSolver(self,category,solvername,input,password):
    """Reenables a solver that has been disabled.

    To provide a new solver, use registerSolver.
    """
    db = None
    try:
      db= NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg = solver.enable(db,category,solvername,input,password)

      # Update web pages
      if msg.find('now enabled') >= 0:
        WebPrinter.makeCatalog(db,self.logger)
        
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg

  def removeSolver(self,category,solvername,input,password):
    """Remove a solver from NEOS.

    If you need to temporarily disable your solver, then you can use
    the disableSolver method instead.
    """

    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      solver = NeosTools.NeosSolver(self.logger)
      msg = solver.removeSolver(db,category,solvername,input,password)

      # Update web pages
      if msg.find('OK') >= 0:
        WebPrinter.makeCatalog(db,self.logger)

      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg
    
  def submitIntermediateResults(self,jobNumber,password,jobOut):
    if password == "test":
      jobsdir=Variables.TESTDIR
    else:
      jobsdir=Variables.JOBSDIR
    db = None
    try:
      db = NeosTools.NeosDB(self.logger,jobsdir)
      job = NeosTools.NeosJob(jobNumber,password,self.logger)
      msg = job.updateResults(db,"job_out",jobOut)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg
    
  def submitFinalResults(self,jobNumber,password,jobResults):
    if password == "test":
      jobsdir=Variables.TESTDIR
    else:
      jobsdir=Variables.JOBSDIR
    db = None
    try:
      db = NeosTools.NeosDB(self.logger,jobsdir)
      job = NeosTools.NeosJob(jobNumber,password,self.logger)
      
      msg = job.updateResults(db,"job_results",jobResults)
      job.printHtml(db)
      db.close()
        
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg

  def submitErrorResults(self,jobNumber,password,errorMessage):
    self.logger.debug("Inside submitErrorResults for job#%d" % jobNumber)

    if password == "test":
      jobsdir=Variables.TESTDIR
    else:
      jobsdir=Variables.JOBSDIR
    db = None
    try:
      db = NeosTools.NeosDB(self.logger,jobsdir)
      job = NeosTools.NeosJob(jobNumber,password,self.logger)

      # set up the email table so an error message is mailed to the contact
      message = "Error with job #%d:\n %s" % (jobNumber,errorMessage)
      result = db.query('SELECT solver_cat,solver_name,solver_input FROM %s WHERE number=%d' % (job.jobsTable,jobNumber))
      if result:
        cat=result[0]
        name=result[1]
        input=result[2]
      
        db.query('insert into EMAIL (job_number,address,message) SELECT %d AS job_number, SOLVERS.contact AS address, "%s" as message FROM SOLVERS WHERE SOLVERS.category="%s" AND SOLVERS.name="%s" AND SOLVERS.input="%s"' % (jobNumber,db.escape_string(message),cat,name,input))
      else:
        self.logger.debug('unknown solver.')

      self.logger.debug('updating job_results')
      msg = job.updateResults(db,"job_results",errorMessage)
      self.logger.debug('printing html')
      job.printHtml(db)
      db.close()
        
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
      
    return msg

  def registerJobPid(self,jobNumber,password,pid):
    db = None
    try:
      db = NeosTools.NeosDB(self.logger)
      job = NeosTools.NeosJob(jobNumber,password,self.logger)
      job.registerPid(db,pid)
      db.close()
    except Exception, (e):
      self.logger.exception(e)
      if db is not None:
        db.close()
      raise Exception, (e)
    return 0


class ThreadingXMLRPCServer(SocketServer.ThreadingMixIn,
                            SimpleXMLRPCServer.SimpleXMLRPCServer):
  pass


class NEOSRequestHandler(SimpleXMLRPCServer.SimpleXMLRPCRequestHandler):
  """
  Use subclass in order to get client information to NeosServer
  """
  def do_POST(self):
    NeosServer._client_address = self.client_address[0]
    SimpleXMLRPCServer.SimpleXMLRPCRequestHandler.do_POST(self)


def isRunning(db):

  result = db.query('SELECT pid,start_time from PROCS WHERE name="NeosServer"')
  pid = 0
  start_time=""
  if result:
    pid = int(result[0])
    start_time = result[1]


    #Check to see if pid is good
    try:
      os.kill(pid,0)

    except OSError, (errno, msg):
      pid = 0
  return (pid,start_time)

if __name__=="__main__":


  db = NeosTools.NeosDB()

  # Check to see if NeosServer already exists
  (pid,startdate) = isRunning(db)
  if pid:
    sys.stdout.write("NeosServer is already running pid=%d, started %s.\n" %
                     (pid,startdate))
    sys.stdout.write("Restart? (y/n) [y]: ")
    line = sys.stdin.readline()
    if line.find('n') == 0:
      sys.exit(1)
    else:
      os.kill(pid,signal.SIGTERM)


  connected = 0
  for i in range(1,100):
    try:
      server=ThreadingXMLRPCServer((Variables.NEOS_HOST, Variables.NEOS_PORT),
                                   NEOSRequestHandler,logRequests=0)
      sys.stdout.write("Started server on port %d\n" % Variables.NEOS_PORT)
      connected=1
      break
    except socket.error, (errno, msg):
      sys.stdout.write(msg + " -- Trying again.\n")
      time.sleep(3)
      
  if not connected:
    sys.stdout.write("Could not bind socket to port %d.\n" % Variables.NEOS_PORT)
    sys.exit(1)
    
  #Daemonize
  pid = os.fork()
  if pid:
    os._exit(0)

  os.setsid() # Set session ID so that we have no controlling terminal



  # Add NeosServer entry to PROCS table
  start_time = time.strftime("%Y-%m-%d %H:%M:%S")
  db.query('REPLACE INTO PROCS (name,pid,start_time) ' +\
          'VALUES("NeosServer","%d","%s")' % (os.getpid(),start_time))
  db.close()
  server.register_instance(NeosServer())
  server.register_introspection_functions()
  try:
    server.serve_forever()
  except Exception, (e):
    logger = logging.getLogger('NeosServer')
    hdlr = logging.FileHandler(os.path.join(Variables.LOGDIR,"NeosServer"))
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr) 
    logger.setLevel(logging.DEBUG)
    logger.debug(str(e))
    
