#!/usr/bin/env python
import sys
import os
import time
import socket
import logging
import logging.handlers
import re
import random
import signal
import xmlrpclib

import NeosXML
from config import Variables
import NeosTools
import WebPrinter

global logger

class Scheduler:
  def __init__(self):
    socket.setdefaulttimeout(5)
    self.neos = xmlrpclib.Server("http://%s:%d"%
                          (Variables.NEOS_HOST,Variables.NEOS_PORT))
    self.db = NeosTools.NeosDB(logger)

  def alrmhandler(self,signo,frame):
    """
    The NEOS Server notifies the scheduler that a job has been submitted by
    sending an alarm signal
    """
    logger.debug("Alarm signal received")

  def runForever(self):
    # Enter into PROCS table
    
    start_time = time.strftime("%Y-%m-%d %H:%M:%S")
    self.db.query('REPLACE INTO PROCS (name,pid,start_time) VALUES("scheduler","%d","%s")' % (os.getpid(),start_time))

    while 1:
      signal.signal(signal.SIGALRM,self.alrmhandler)
      time.sleep(30) # Will get woken up by SIGALRM when job is put in queue
      signal.signal(signal.SIGALRM,signal.SIG_IGN)
       
      results = self.db.queryAll('SELECT QUEUE.job_number,QUEUE.priority,QUEUE.station_id,JOBS.solver_cat,JOBS.solver_name,JOBS.solver_input,JOBS.password FROM QUEUE LEFT JOIN JOBS ON QUEUE.job_number=JOBS.number WHERE QUEUE.status="Waiting" AND QUEUE.priority!="test" ORDER BY QUEUE.job_number',logme=False)

      logger.debug(str(results))
      if len(results) == 0:
        #Check for any test jobs
        results = self.db.queryAll('SELECT QUEUE.job_number,QUEUE.priority,QUEUE.station_id,TESTJOBS.solver_cat,TESTJOBS.solver_name,TESTJOBS.solver_input,TESTJOBS.password FROM QUEUE LEFT JOIN TESTJOBS ON QUEUE.job_number=TESTJOBS.number WHERE QUEUE.status="Waiting" ORDER BY QUEUE.job_number',logme=False)

      #if len(results) == 0:
        #logger.debug("No jobs ... sleeping")

      for result in results:
        logger.debug(str(result))
        #For each job in the queue...
        logger.debug("Trying to schedule job #%d" % result[0])

        # Get the job's xml file and solver name
        jobNumber = result[0]
        priority = result[1]
        station_id = result[2]
        category = result[3]
        solver = result[4]
        input = result[5]
        fullSolverName=":".join(result[3:6])
        password = result[6]
        
        jobinSize = self.db.getJobInSize(jobNumber)
        if jobinSize > Variables.MAX_FILE_LENGTH:
          self.terminate(jobNumber,password,"JobIn file length exceeds maximum of %d" % Variables.MAX_FILE_LENGTH)
          continue
        
        jobIn = self.db.getJobIn(jobNumber)
          
        # get solver xml
        result = self.db.query('SELECT xmldescription,status,contact FROM SOLVERS where category="%s" and name="%s" and input="%s"' % (category,solver,input))
        
        if result and len(result):
          solverXml = result[0]
          status = result[1]
          contact = result[2]
          if status == "Disabled":
            self.terminate(jobNumber,password,"Solver %s:%s:%s has been disabled.  Contact %s for details." % (category,solver,input,contact))
            continue
        else:
          self.terminate(jobNumber,password,
                         "Solver %s:%s:%s is unknown"%(category,solver,input))
          continue
        
        # Find out which station to run on
        # If used for testing, then the job might have
        # a station already designated
        self.errorMessage = None
        try:
          if station_id is None or station_id==0:
            station_id = self.selectStation(priority, fullSolverName)

          if station_id == 0 and not self.wait:
            if self.errorMessage:
              self.terminate(jobNumber,password,"%s.\n Please contact solver administrator %s" % (self.errorMessage,contact))
            else:
              self.terminate(jobNumber,password,"No station found running solver %s.\n  Please contact solver administrator %s" % (fullSolverName,contact))
            continue
        except Exception, (error):
          self.terminate(jobNumber,password,"Exception selecting station: " + str(error))
          continue

        # Schedule the job
        if station_id:
          try:
            self.scheduleJob(jobNumber,password,station_id, priority,
                             jobIn, solverXml)
          except Exception, (error):
            self.terminate(jobNumber,password,"Exception scheduling job: " + str(error))
            continue
        
        else:
          logger.debug("Could not schedule job #%d" % jobNumber)

  
    
  def selectStation(self, priority, fullSolverName):
    # Get list of stations for this solver and pick one
    (category,solver,inputMethod) = fullSolverName.split(':')
    result = self.db.query('SELECT stations FROM SOLVERS where category="%s" and name="%s" and input="%s"' % (category,solver,inputMethod))

    #stations is '|'-delimited string of numbers
    logger.debug(result[0])
    station_list = result[0].split('|')
    logger.info("station_list = %s" % str(station_list))

    
    stationload = {}
    for station in station_list:
      stationload[int(station)]=0

    #Find which stations are being used
    results= self.db.queryAll('SELECT station_id FROM QUEUE WHERE status="Running"')
    for (station_id,) in results:
      if station_id in station_list:
        stationload[station_id]+=1

    #Delete stations that are maxed out or not right priority
    responsive_station_list=[]
    self.wait=0
    for station in station_list:
      result = self.db.query("SELECT priority,max_jobs,host_name,port,user FROM STATIONS WHERE ID=%s" % station)
      station_priority = result[0]
      max_jobs = result[1]
      host_name = result[2]
      port = result[3]
      user = result[4]
      
      if priority == "long" and station_priority == "short":
        logger.debug("Removing station %s because of wrong priority" % station)
        continue
      elif stationload[int(station)] >= max_jobs:
        logger.debug("Removing station %s because it's maxed out" % station)
        # keep in queue
        self.wait=1
        continue

      try:
        reply = self.neos.pingHost(user,host_name)
        if reply != "OK":
          self.errorMessage = reply
        else:
          pinger = xmlrpclib.Server("http://%s:%d" % (host_name,port))
          reply = pinger.verifySolver(fullSolverName)
          logger.debug("pinger verified http://%s:%d - %s" % (host_name,port,reply))
          if reply != "OK":
            self.errorMessage = reply

      except Exception, e:
        reply = ""
        logger.debug("pinger could not verify http://%s:%d" %(host_name,port))
        logger.debug(str(e))
      if reply.find("OK") < 0:
        logger.debug("Removing station %s because it can't be reached" % station)
      else:
        responsive_station_list.append(station)
  

    # Of the remaining stations, find stations with minimum load
    logger.debug("stationload: %s" % str(stationload))
    if len(responsive_station_list):
      leastUsedStationsList = [int(responsive_station_list[0])]
      for station in responsive_station_list[1:]:
        if stationload[int(station)] < stationload[leastUsedStationsList[0]]:
          leastUsedStationsList = [int(station)]
        elif stationload[int(station)] == stationload[leastUsedStationsList[0]]:
          leastUsedStationsList.append(int(station))

      # If there is a tie, pick one at random
      logger.debug("least list: " + str(leastUsedStationsList))
      return random.choice(leastUsedStationsList)
    else:
      return 0


  def scheduleJob(self,jobNumber,password,station_id, priority,
                  jobXml,solverXml):
    
    self.db.query('UPDATE QUEUE SET station_id="%d" WHERE job_number="%d"' %(station_id, jobNumber))

    
    r = self.db.query('SELECT solver_cat,solver_name,solver_input,host_name,max_jobs,port,pid,time_limit,mem_limit,file_limit,filesize_limit FROM STATIONS WHERE ID=%d' % station_id)

    if not r:
      logger.error("STATION ID %d not in STATIONS table" %(station_id))
      self.terminate(jobNumber,password,
                     "Error in database, Job assigned to unknown station %d"
                     % (station_id))
      return
    solver = ":".join(r[0:3])
    station=r[3]
    maxjobs = r[4]
    port = int(r[5])
    pid = int(r[6])
    limits = {} 
    limits['time_limit'] = r[7]
    limits['mem_limit'] = r[8]
    limits['file_limit'] = r[9]
    limits['filesize_limit'] = r[10]
    limits['solver'] = solver
    if not port or not pid:
      self.terminate(jobNumber,password,
                 "no port/pid registered on %s for solver %s"%(station,solver))
      return

    logger.info("Trying to connect to http://%s:%d" % (station,port))
    try:
      sock = xmlrpclib.Server("http://%s:%d" % (station,port))
      msg = sock.ping()
      if msg != "OK":
        self.terminate(jobNumber,password,"pinging host failed: " + msg)
        return
    except socket.error, (error):
      self.terminate(jobNumber,password,"Exception pinging host %s: %s" %(station,str(error)))
      return
    except xmlrpclib.Fault, fault:
      self.terminate(jobNumber,password,fault.faultString)
      return

    logger.info("Sending job %d to %s:%d (pid=%d)",
                jobNumber,station,port,pid)


    logger.info("Updating queue status, starttime for job %d",jobNumber)
    self.db.query('UPDATE QUEUE SET status="Running" WHERE job_number=%d'%(jobNumber))
    starttime=time.strftime("%Y-%m-%d %H:%M:%S")
    if priority=="test":
      self.db.query('UPDATE TESTJOBS SET start_time="%s",station_id=%d,status="Running" where number=%d'%(starttime,station_id,jobNumber))
    else:
      self.db.query('UPDATE JOBS SET start_time="%s",station_id=%d,status="Running" where number=%d'%(starttime,station_id,jobNumber))

    # Create a results page
    if not os.access(os.path.join(Variables.NEOS_WWW_HOME,"jobs"),os.X_OK):
      os.mkdir(os.path.join(Variables.NEOS_WWW_HOME,"jobs"))
    webpage=file(os.path.join(Variables.NEOS_WWW_HOME,"jobs","%d.html"%jobNumber),"w")
    WebPrinter.printHtmlHeader(webpage,jobNumber,solver,station,starttime)
    webpage.write("No results yet.  Try reloading page later.\n")
    webpage.close()
    
    self.neos.submitIntermediateResults(jobNumber,password,
                            "Job %d sent to %s\n" % (jobNumber,station))

    try:
      logger.debug("making tarball")
      tarballbin=NeosXML.makeTarball(jobXml,solverXml,limits,logger)
      tarball64=xmlrpclib.Binary(tarballbin)
    except Exception, (err):
      self.terminate(jobNumber,password,"Exception making tarball:" +str(err))
      return

    try:
      logger.debug("sending tarball")
      oldtimeout = socket.getdefaulttimeout()
      socket.setdefaulttimeout(600)
      retcode = sock.beginJob(jobNumber,password,solver,tarball64)
      socket.setdefaulttimeout(oldtimeout)
    except Exception, (err):
      self.terminate(jobNumber,password,"Exception sending tarball:" +str(err))
      return


    

  def isRunning(self):
    db = NeosTools.NeosDB()
    result = self.db.query('SELECT pid,start_time from PROCS WHERE name="scheduler"')
    db.close()
    pid = 0
    start_time=""
    if result:
      pid = int(result[0])
      start_time = result[1]
      logger.info("scheduler pid=%d started %s",pid,start_time)
            

      #Check to see if pid is good
      try:
        os.kill(pid,0)
        logger.warning("Process #%d is running.",pid)

      except OSError, (errno, msg):
        pid = 0

    return (pid,start_time)

  def makePageResults(self,jobNumber,solver,starttime,station):
    pagename=os.path.join(Variables.NEOS_WWW_HOME,jobs,"%d.html" % jobNumber)
    page = file(pagename,"w")
    page.write("""
    """ % (jobNumber, solver, starttime, station))
        
        
  def terminate(self, jobNumber, password, msg):
    """
    If there is an error scheduling job, send termination message
    to NEOS.
    """
    msg += "\n"
    logger.warning("Job#%d: %s" %(jobNumber,msg))
    try:
      logger.debug("submitErrorResults to NEOS server for job#%d " % jobNumber)
      info = self.neos.submitErrorResults(jobNumber,password,msg)
      logger.debug("Done submitErrorResults to NEOS server for job#%d " % jobNumber)

    except Exception, (err):
      logger.exception(err)



if __name__=="__main__":
  if not os.access(Variables.LOGDIR, os.X_OK):
    os.makedirs(Variables.LOGDIR)
  logger = logging.getLogger('scheduler')
  hdlr = logging.handlers.RotatingFileHandler(os.path.join(Variables.LOGDIR,"scheduler"),maxBytes=100000,backupCount=1)
  formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
  hdlr.setFormatter(formatter)
  logger.addHandler(hdlr)
  logger.setLevel(logging.DEBUG)

  scheduler = Scheduler()  
  # Check to see if scheduler already exists
  (pid,startdate) = scheduler.isRunning()
  if pid:
    sys.stdout.write("Scheduler is already running pid=%d, started %s.\n" %
                     (pid,startdate))
    sys.stdout.write("Restart? (y/n) [y]: ")

    line = sys.stdin.readline()
    if line.find('n') == 0:
      sys.exit(1)
    else:
      os.kill(pid,signal.SIGTERM)

    


  logger.info("Started scheduler")

  # daemonize
  pid = os.fork()
  if pid:
    os._exit(0)

  os.setsid() # Set session ID so that we have no controlling terminal
  os.umask(033) # fix creation mask
  for i in range(3):
    try:
      os.close(i)
    except OSError, e:
      if e.errno != errno.EBADF:
        raise RuntimeError('Could not close default descriptor ' + str(i))

  try:
    scheduler.runForever()
  except Exception, (e):
    logger.exception(e)
    
  

