#!/usr/bin/env python
import socket
import SocketServer
import re
import sys
import os
import signal
import shutil
import popen2
import time
import select
import xmlrpclib
import threading
import SimpleXMLRPCServer

from config import Variables

def getDriver(solver):
  if not os.access(Variables.DRIVER_FILE,os.R_OK):
    return None
  driverFile =  open(Variables.DRIVER_FILE,"r")
  for line in driverFile.readlines():
    m = re.match(solver+r'\s+(\S+)',line,re.IGNORECASE)
    if m and len(m.groups()):
      return m.groups()[0]
  return None
      

class SolverDaemon:
  """
  daemonServer Runs solver for NEOS

  Recognized request:
  verify()
    returns "OK"
  ping()
    returns "OK"
    
  verifyJob(num, pid)
    returns ("OK" or "Job <num> Not Running")

  verifySolver(solver)
    returns ("OK" if solver will run, msg otherwise)
    
  beginJob(num,pass,solver,tarball)

  killJob(num,pass,pid)
  """
  def __init__(self):
    self.neos = xmlrpclib.ServerProxy("http://%s:%d" %
                                 (Variables.NEOS_HOST,Variables.NEOS_PORT))

    # Start logging if available.  
    # logging introduced in python 2.3
    if not os.access(Variables.LOGDIR,os.X_OK):
      sys.stdout.write("Directory %s does not exist...Creating.\n"%Variables.LOGDIR)
      sys.stdout.flush()
      os.makedirs(Variables.LOGDIR)
      
    try:
      import logging
      import logging.handlers
      self.logger = logging.getLogger('SolverDaemon')
      hdlr = logging.handlers.RotatingFileHandler(os.path.join(Variables.LOGDIR,"SolverDaemon"),maxBytes=100000,backupCount=1)
      formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
      hdlr.setFormatter(formatter)
      self.logger.addHandler(hdlr) 
      self.logger.setLevel(logging.DEBUG)
      
    except ImportError, e:
      sys.stderr.write("Your python version does not support limited-file-size logging, an update to 2.3 is suggested but not required.\n")
      self.logger = DefaultLogger(os.path.join(Variables.LOGDIR,"SolverDaemon"))
      

      

  def killJob(self,jobNumber,passwd,pid):
    self.logger.info("Request to kill job # %d " % jobNumber)
    if os.name == "posix":
      try:
        self.neos.submitIntermediateResults(jobNumber,passwd,"---Killed by request.---")
        os.kill(pid,signal.SIGKILL)
        self.logger.info("Job %d Killed by Request" % jobNumber)
        return "Job %d Killed" % jobNumber

      except OSError, (errno, msg):
        if errno==3: # No such process
          self.logger.info("Could not kill job %d -- not running" % jobNumber)
          return("Job %d Not Running" % jobNumber)

        else:
          self.logger.info("Could not kill job %d" % jobNumber)
          return "Could not kill job %d" % jobNumber
    else:
      return "Could not kill -- windows"

  def beginTest(self,solver,tarball):
    self.logger.info("Testing solver %s" % solver)
    jobdir = os.path.join(Variables.TESTDIR,solver)
    if os.access(jobdir,os.X_OK):
      for f in os.listdir(jobdir):
        os.remove(os.path.join(jobdir,f))
      os.rmdir(jobdir)

    os.makedirs(jobdir)
    os.chdir(jobdir)

    if self.extractTarball(tarball):
      self.neos.submitIntermediateResults(0,solver,"Extracted tar files.\n")
    else:
      self.neos.submitIntermediateResults(0,solver,"Error: Could not extract tar files.")
      return -1
    self.neos.submitIntermediateResults(0,solver,"---------- Begin Solver Output -----------\n")
    

    executer = ExecutionThread(0,"",solver,self.neos,self.logger)
    executer.setDaemon(True)
    executer.start()
    return 0

  def beginJob(self, jobNumber, password, solver, tarball):
    try:
      if password=="test":
        self.logger.info("Request to start testing job # %d (%s)" % (jobNumber,solver))
        jobdir = os.path.join(Variables.TESTDIR,str(jobNumber))
      else:
        self.logger.info("Request to start job # %d (%s)" % (jobNumber,solver))
        jobdir = os.path.join(Variables.JOBSDIR,str(jobNumber))

      # Create directory (empty if necessary) to run job in and cd to it,
      self.solver = solver
      if os.access(jobdir,os.X_OK):
        shutil.rmtree(jobdir)
      os.makedirs(jobdir)
      os.chdir(jobdir)
    except Exception, e:
      self.logger.exception(e)
      self.neos.submitErrorResult(jobNumber,password,str(e))
      os._exit(1)
    
    try:
      self.neos.submitIntermediateResults(jobNumber, password,"jobnumber=%d, host=%s\n" % (jobNumber,socket.getfqdn()))
    
      if self.extractTarball(tarball):
        self.neos.submitIntermediateResults(jobNumber,password,"Extracted tar files.\n")
      else:
        self.neos.submitIntermediateResults(jobNumber,password,"Error: Could not extract tar files.")
        return -1
      self.neos.submitIntermediateResults(jobNumber,password,"---------- Begin Solver Output -----------\n")
      if os.name == "nt":
        executer = WindowsExecutionThread(jobNumber,password,self.solver,self.neos,self.logger)
      else:
        executer = ExecutionThread(jobNumber,password,self.solver,self.neos,self.logger)
      executer.setDaemon(True)
      executer.start()
    except Exception, (e):
      self.logger.exception(e)
      raise e
    return 0


  def extractTarball(self,tarball):
    f = open("jobin.tar","wb")
    f.write(tarball.data)
    f.close()
    self.logger.debug("tarball written.")
    try:
      import tarfile
      tar = tarfile.open("jobin.tar","r")
      for member in tar.getmembers():
        tar.extract(member)
      tar.close()
    except ImportError, (error):
      self.logger.info("Could not import tarfile module, using system tar")
      os.system("tar -xf jobin.tar")
    except Exception, (error):
      self.logger.exception(error)
      return False
    return True
    

  def ping(self):
    self.logger.info("ping request.")
    return "OK"

  def verifyJob(self,jobNumber,pid):
    self.logger.info("verifyJob request")
    try:
      os.kill(pid,0)
      return "OK"
    except OSError, (errno, msg):
      return("Job %d Not Running" % jobNumber)

  def verifySolver(self,solver):
    """
    returns ("OK" if solver will run, msg otherwise)
    """
    if not hasattr(Variables,"DRIVER_FILE") or not os.access(Variables.DRIVER_FILE,os.R_OK):
      return "No driver file found on server"
    driver = getDriver(solver)
    if driver:
      return "OK"
    else:
      return "No driver for solver %s on %s@%s" % (solver,os.getenv('LOGNAME'),
                                     socket.getfqdn(socket.gethostname()))
    


class Listener(threading.Thread):
  """
  Listener is defined to provide nonblocking reads of stdout and stderr
  from the driver program.  It is only used when select is not available
  (Windows)
  """
  def __init__(self, f, jobNumber, password, neos,logger):
    threading.Thread.__init__(self)
    self.f = f
    self.jobNumber = jobNumber
    self.password = password
    self.neos = neos
    self.logger = logger

  def run(self):
    self.stop = 0
    while not self.stop:
      try:
        msg = self.f.readline()
        if msg:
          self.neos.submitIntermediateResults(self.jobNumber,
                                              self.password,
                                             xmlrpclib.Binary(msg))
      except IOError, (err):
        if err[0] == 4:
          self.logger.info("Child process dead, exiting")
        else:
          self.logger.exception(err)
        self.stop=1
      except Exception, (e):
        self.logger.exception(e)
        self.stop=1
    
class WindowsExecutionThread(threading.Thread):
  def __init__(self,jobnumber,password,solver,neos,logger):
    threading.Thread.__init__(self)
    self.neos=neos
    self.logger=logger
    self.jobNumber = jobnumber
    self.password = password
    self.solver = solver

  def run(self):
    ret     = None
    out     = ''
    err     = ''

    if not hasattr(Variables,"DRIVER_FILE") or not os.access(Variables.DRIVER_FILE,os.R_OK):
      self.neos.submitIntermediateResults(self.jobNumber,self.password,
                                          "File %s not readable.\n" %
                                          Variables.DRIVER_FILE)
                                          
    try:
      driver = getDriver(self.solver)
    except Exception, e:
      driver = None
    
    if not driver:
      self.neos.submitFinalResults(self.jobNumber,self.password,
                                   "No driver for solver %s on %s@%s"%
                                   (self.solver,os.getenv('LOGNAME'),
                                    socket.getfqdn(socket.gethostname())))
      self.logger.error("No driver for solver %s on %s@%s"%
                        (self.solver,os.getenv('LOGNAME'),
                         socket.getfqdn(socket.gethostname())))
      return
      
    
    if os.access("neos-time_limit",os.R_OK):
      f = open("neos-time_limit","r")
      timeLimit = long(f.read())
      f.close()
    else:
      timeLimit = 1000000L
      
    self.logger.info("Executing %s" % (driver))
    (output,input,error) = popen2.popen3(driver)
    input.close()
    outputClosed = 0
    errorClosed  = 0
    outputListener = Listener(output,self.jobNumber,self.password,self.neos,
                              self.logger)
    errorListener = Listener(error,self.jobNumber,self.password,self.neos,
                             self.logger)
    outputListener.start()
    errorListener.start()
    outputListener.join(timeLimit)
    errorListener.join(10)

    self.logger.debug("Sending final results..")

    #return job.results if it exists
    if os.access("job.results",os.R_OK):
      results = open("job.results","rb")
      self.neos.submitFinalResults(self.jobNumber,self.password,xmlrpclib.Binary(results.read()))
      results.close()
    else:
      self.neos.submitFinalResults(self.jobNumber,self.password,"")

    return


    
class ExecutionThread(threading.Thread):

  def __init__(self,jobnumber,password,solver,neos,logger):
    threading.Thread.__init__(self)
    self.neos=neos
    self.logger=logger
    self.jobNumber = jobnumber
    self.password = password
    self.solver = solver

    
  def run(self):
    ret     = None
    out     = ''
    err     = ''
    try:
      self.logger.debug("self.jobNumber=%d" % self.jobNumber)
      self.logger.debug("self.password=%s" % self.password)
      self.logger.debug("os.getpid =%d" % os.getpid())
    except Exception, e:
      self.logger.exception(e)
      self.neos.submitErrorResults(self.jobNumber,self.password,str(e))
      os._exit(1)

    
    #only does something for posix systems
    try:
      self.setrlimits()
    except Exception, e:
      self.neos.submitErrorResults(self.jobNumber,self.password,
                                   "Error reading files: %s\n" % (str(e)))
      self.logger.error("Error reading files: %s" % (str(e)))
      return
                        

    if not hasattr(Variables,"DRIVER_FILE") or not os.access(Variables.DRIVER_FILE,os.R_OK):
      self.neos.submitErrorResults(self.jobNumber,self.password,
                              "DRIVER_FILE in config file not readable.\n")
      self.logger.error("Error: DRIVER_FILE not readable or not set in config file")
      return
                                          
    try:
      driver = getDriver(self.solver)
    except Exception, e:
      driver = None
    
    if not driver:
      self.neos.submitErrorResults(self.jobNumber,self.password,
                                   "No driver for solver %s on %s@%s"%
                                   (self.solver,os.getenv('LOGNAME'),
                                    socket.getfqdn(socket.gethostname())))
      self.logger.error("No driver for solver %s on %s@%s"%
                        (self.solver,os.getenv('LOGNAME'),
                         socket.getfqdn(socket.gethostname())))
      return
      

    if os.access("neos-time_limit",os.R_OK):
      f = open("neos-time_limit","r")
      timeLimit = long(f.read())
      f.close()
      self.logger.debug("Setting alarm signal catcher")
      self.logger.debug("Setting alarm for %d seconds." % timeLimit)
      self.timer = threading.Timer(timeLimit, self.alarmcatcher, args=[self.jobNumber])
      self.timer.start()

    self.logger.info("Executing %s" % (driver))
    
    pipe = popen2.Popen3(driver,1)
    self.neos.registerJobPid(self.jobNumber,self.password,pipe.pid)
    self.child = pipe.pid
    self.logger.debug("Child is pid %d" % self.child)
    self.logger.debug("I am pid %d" % os.getpid())
    input = pipe.tochild
    output = pipe.fromchild
    error = pipe.childerr

    input.close()
    outputClosed = 0
    errorClosed  = 0
    self.stop=0

    while not self.stop:
      try:
        ready = select.select([output, error], [], [])
        if len(ready[0]):
          if error in ready[0]:
            msg = error.readline()
            if msg:
              self.neos.submitIntermediateResults(self.jobNumber,
                                                  self.password,
                                        xmlrpclib.Binary("Error: " + msg))
            else:
              errorClosed = 1
          if output in ready[0]:
            msg = output.readline()
            if msg:
              self.neos.submitIntermediateResults(self.jobNumber,
                                                  self.password,
                                                  xmlrpclib.Binary(msg))
            else:
              outputClosed = 1
        if outputClosed and errorClosed:
          self.stop=1
      except IOError, (err):
        if err[0] == 4:
          self.logger.info("Child process dead, exiting")
        else:
          self.logger.exception(err)
        self.stop=1
      except Exception, (e):
        self.logger.exception(e)
        self.stop=1
    if hasattr(self,'timer'):
      self.timer.cancel()

    try:
      self.logger.debug("Closing output,error")
      output.close()
      error.close()
    except Exception, (e):
      self.logger.debug(e)

    self.logger.debug("Sending final results..")


    #return job.results if it exists
    if os.access("job.results",os.R_OK):
      results = open("job.results","rb")
      self.neos.submitFinalResults(self.jobNumber,self.password,xmlrpclib.Binary(results.read()))
      results.close()
    else:
      self.neos.submitFinalResults(self.jobNumber,self.password,"")

    return

  def alarmcatcher(self, jobNumber):
    self.logger.debug("Alarm tripped on job %d." % jobNumber)
    self.logger.debug("Sending kill to pid %d." % self.child)
    self.neos.submitFinalResults(jobNumber,self.password,"Time limit reached. Process killed\n")
    try:
      os.kill(self.child,signal.SIGKILL)
    except OSError, (err):
      if err[0] != 3:
        self.logger.exception(err)
    except Exception, (err):
      self.logger.exception(err)

    self.stop=1


  def setrlimits(self):
    if os.name == "posix":
      import resource
      #get tuples (soft,hard) for limit values
      r_numfiles = resource.getrlimit(resource.RLIMIT_NOFILE)
      r_memlimit = resource.getrlimit(resource.RLIMIT_AS)
      r_filesize = resource.getrlimit(resource.RLIMIT_FSIZE)
      self.logger.debug("got rlimits")

      if os.access("neos-mem_limit",os.R_OK):
        f = open("neos-mem_limit","r")
        memLimit=long(f.read())
        f.close()
        if memLimit < r_memlimit[0]:
          resource.setrlimit(resource.RLIMIT_AS,(memLimit,r_memlimit[1]))
      self.logger.debug("set rss slimit")
      if os.access("neos-file_limit",os.R_OK):
        f = open("neos-file_limit","r")
        fileLimit=long(f.read())
        f.close()
        if fileLimit < r_numfiles[0]:
          resource.setrlimit(resource.RLIMIT_NOFILE,(fileLimit,r_numfiles[1]))
      self.logger.debug("set nofiles slimit")
      if os.access("neos-filesize_limit",os.R_OK):
        f = open("neos-filesize_limit","r")
        fileSizeLimit=long(f.read())
        f.close()
        if fileSizeLimit < r_numfiles[0]:
          resource.setrlimit(resource.RLIMIT_FSIZE,(fileSizeLimit,r_filesize[1]))
      self.logger.debug("set filesize slimit")

    else:
      pass


class DefaultLogger:
  """
  Used for compatibility with python versions < 2.3
  """
  def __init__(self,filename):
    self.log=open(filename,"a")

  def error(self,msg):
    self.log.write(time.asctime() + " ERROR " + msg + "\n")
    self.log.flush()

  def warning(self,msg):
    self.log.write(time.asctime() + " WARNING " + msg + "\n")
    self.log.flush()

  def debug(self,msg):
    self.log.write(time.asctime() + " DEBUG " + msg+ "\n")
    self.log.flush()

  def info(self,msg):
    self.log.write(time.asctime() + " INFO " + msg + "\n")
    self.log.flush()

  def critical(self,msg):
    self.log.write(time.asctime() + " CRITICAL " + msg + "\n")
    self.log.flush()

  def exception(self,err):
    self.log.write(time.asctime() + " EXCEPTION " + str(err) + "\n")
    self.log.flush()


class ThreadingXMLRPCServer(SocketServer.ThreadingMixIn,
                            SimpleXMLRPCServer.SimpleXMLRPCServer):
  def verify_request(self,request, client_address):
    """
    Only recognize requests coming from NEOS or localhost
    """
    return ( client_address[0] == socket.gethostbyname(Variables.NEOS_HOST)
           or client_address[0] == socket.gethostbyname(socket.gethostname()))


  
def checkForExistingDaemon():
  neos = xmlrpclib.Server("http://%s:%d" %
                          (Variables.NEOS_HOST, Variables.NEOS_PORT))
  response = neos.listSolversOnStation(os.getenv('LOGNAME'),
                              socket.getfqdn(socket.gethostname()))
  daemonrunning=0
  if len(response):
    sys.stdout.write("The following solvers are registered on this user/machine:\n")
    for s in response:
      m = re.match(r'(\S+)\s+(\d+)\s+(\d+)',s)
      if m and m.groups() and len(m.groups()) == 3:
        solver = m.groups()[0]
        port = int(m.groups()[1])
        pid = int(m.groups()[2])
        sys.stdout.write("Solver: %s port: %d pid: %d\n"%(solver,port,pid))
        if pid:
          daemonrunning = pid
      else:
        sys.stdout.write("Received from NEOS: " + s + "\n")
        sys.exit(1)
  else:
    sys.stdout.write("There is no solver registered on NEOS for %s@%s\n" %
                     (os.getenv('LOGNAME'),
                      socket.getfqdn(socket.gethostname())))
    sys.exit(1)
                     

  if daemonrunning:
    sys.stdout.write("NEOS believes there is already a daemon running on your machine\n")
    try:
      os.kill(daemonrunning,0)
      sys.stdout.write("There is a process of that number running locally.\n")
      sys.stdout.write("Continue? (will kill old daemon) (y/n) [n]:")
      sys.stdout.flush()
      line = sys.stdin.read(1)
    except OSError, (errno, msg):
      if errno == 3:
        line = "y"
      else:
        raise OSError(errno, msg)
        
      
    if line.lower() == "y" or line.lower() == "yes":
      try:
        os.kill(daemonrunning,signal.SIGKILL)
        sys.stdout.write("SIGKILL sent to process %d\n" % daemonrunning)
      except OSError, (errno, msg):
        if errno == 3:
          sys.stdout.write("process %d did not exist.\n" % daemonrunning)
        else:
          raise OSError(errno, msg)
    else:
      sys.stdout.write("Quitting.\n")
      sys.exit(0)


if __name__=="__main__":
  if os.name == "nt":
    username = os.getenv('USERNAME')
  else:
    username = os.getenv('LOGNAME')
    # Don't know how to check for daemon in windows
    checkForExistingDaemon()

  if len(sys.argv) >= 2:
    try:
      port = int(sys.argv[1])
      server = ThreadingXMLRPCServer((socket.gethostname(),port),logRequests=0)
      
    except ValueError, (e):
      sys.stdout.write("Usage: SolverDaemon [port #]\n")
      sys.stdout.write("Default port: first unused between 4000-4999\n")
      sys.exit(1)
    except socket.error, (errno, msg):
      sys.stdout.write("Port %d in use.  Quitting.\n")
      sys.exit(1)

  else:
    for port in range(4000,5000):
      try:
        server = ThreadingXMLRPCServer((socket.gethostname(),port),logRequests=0)
        break
      except socket.error, (errno, msg):
        if errno != 98 and errno != 125 and errno != 10048: 
          # already in use, linux, sol, win
          raise socket.error(errno, msg)
        if port == 4999:
          sys.exit(1)

  if os.name == "posix":
    pid = os.fork()
    if pid:
      os._exit(0)

    pid = os.getpid()
    os.setsid()
    os.umask(077)

  else:
    pid = 1
    
  neos = xmlrpclib.Server("http://%s:%d"%(Variables.NEOS_HOST,Variables.NEOS_PORT))
  response = neos.registerDaemon(username,
                                 socket.getfqdn(socket.gethostname()),port,pid)
  sys.stdout.write(response)
  
  
  try:
    server.register_instance(SolverDaemon())
    server.serve_forever()
  except Exception, (err):
    sys.stdout.write("oops." + str(err) + "\n")
